//	<><><><><><><><><><><><><>  Plot3DT.h  <><><><><><><><><><><><><><> 
//
// -----------------------------------------------------------------
// (C) Copyright Clark Dailey, Inc. 2006
//
// Plot3DT is a base class for triangle mesh plot
//
// -----------------------------------------------------------------

#ifndef	_PLOT3DT_H_	 	// avoid repeated including
#define	_PLOT3DT_H_	1	// has been included

#include "plot3d.h"

// ----------------
//   point class
// ----------------
class Point3DT
{
public:
	Point3DT() { m_is_visible=-1; }
	~Point3DT() { }
	void Load(unsigned i,unsigned j,double x,double y,double z,double r,double g,double b);
	void Project(double xp,double yp,double zp) { m_xp=xp; m_yp=yp, m_zp=zp; }

public: // data
	unsigned m_i,m_j;		// mesh index
	double m_xf,m_yf,m_zf;	// function coordinates
	double m_xn,m_yn,m_zn;	// normal surface vector
	double m_xp,m_yp,m_zp;	// projection coordinates
	double m_red, m_green, m_blue;	// color of point
	int m_is_visible;		// -1=unknown, 0=no, 1=yes
};

// ----------------
//  triangle class
// ----------------
class Triangle3DT
{
public:
	Triangle3DT() { m_pPnt[0] = m_pPnt[1] = m_pPnt[2] = 0; m_color=PLOT_COLOR_WHITE; }
	~Triangle3DT() { }
	// get coordinates of triangle
	double XPnt(unsigned i) { return(m_pPnt[i]->m_xp); }
	double YPnt(unsigned i) { return(m_pPnt[i]->m_yp); }
	double ZPnt(unsigned i) { return(m_pPnt[i]->m_zp); }
	// get color of points
	double RedPnt  (unsigned i) { return(m_pPnt[i]->m_red);   }
	double GreenPnt(unsigned i) { return(m_pPnt[i]->m_green); }
	double BluePnt (unsigned i) { return(m_pPnt[i]->m_blue);  }
	double RedTri  () { return(m_red);   }
	double GreenTri() { return(m_green); }
	double BlueTri () { return(m_blue);  }

	// returns: 0=no, 1=point is one of defining points of triangle
	int IsDefiningPoint(Point3DT* pPnt);

		// is a given xp,yp point within the triangle
	// returns: 0=no, 1=point is within triangle
	int IsInscribed(double xp,double yp);

	// get color of triangle
	long GetColor() { return(m_color); }

	// calculate zp intercept at a given xp,yp point
	double ZIntercept(double xp,double yp);

	void Load(unsigned i,unsigned j,unsigned k,Point3DT* pP1,Point3DT* pP2,Point3DT* pP3);
	void SetColor(double r,double g,double b);

public: // data
	Point3DT* m_pPnt[3]; // pointer to three points (we don't own the memory)
	unsigned m_i,m_j,m_k;	// indexes of triangle
	double m_xptmin,m_xptmax,m_yptmin,m_yptmax,m_zptmin,m_zptmax; // min and max projections
	double m_dxp[3],m_dyp[3],m_dzp[3];	// point-to-point deltas
	double m_A, m_B, m_C, m_D;	// plane constants (A*xp+B*yp+C*zp+D=0)
	double m_area; // area of triangle
	double m_dir;  // direction of vertices (-1 or +1)
	long   m_color;	// triangle color (Windows)
	double m_red; 	// 0-100
	double m_green;	// 0-100
	double m_blue; 	// 0-100
};


// --------------------------------
//  3D triangle mesh plotting class
// --------------------------------
class Plot3DT : public Plot3D
{
public:
	Plot3DT();
	virtual ~Plot3DT();

	int ExportPOV (CString  povFilePath,CString sPovObjectName,  int is_smooth,int is_color);
	int ExportVRML(CString vrmlFilePath,CString sVrmlSurfaceName,int isColor,int isBack,int drawLine1,int drawLine2);
	int ExportDXF (CString sDxfFilePath,CString sDxfSurfaceName);

protected:	// routines
	// load / save
	int ReadFile(void);
	int WriteFile(void);

	// plotting
	void ResetT(void);

	// retrieve triangle info
	unsigned TriangleCount() { return(m_nTriangles); }
	Triangle3DT* GetTriangle(unsigned index) { return( (index<m_nTriangles) ? &m_pTriangles[index] : (Triangle3DT*)0 ); }

	// calculate coordinates for export to Persistence-of-Vision (POV)
	void GetPOVCamera(double* xc,double* yc,double* zc);

	virtual double XParametric(double u,double v)=0;
	virtual double YParametric(double u,double v)=0;
	virtual double ZParametric(double u,double v)=0;
	virtual int InitEquations(void)=0;

	// color info
	virtual double RedColor  (double u,double v,double w)=0;
	virtual double GreenColor(double u,double v,double w)=0;
	virtual double BlueColor (double u,double v,double w)=0;

	int  InitPlot(void);
	int  CalcPoints(void);
	int  CalcTriangles(void);
	int  PlotColorPixels(void);
	int  DrawULines(void);
	int  DrawVLines(void);
	void DefInputsT(void);
	void FreeMemory(void);
	int  PlotLine(int FirstLine,Point3DT* pPnt1,Point3DT* pPnt2);
	long CalcPixelColor(int xv,int yv);
	int  IsPointVisible(double xp,double yp,double zp);
	int  IsPointVisible(Point3DT* pPnt);
	Point3DT* Point(unsigned i,unsigned j);
	unsigned HashIndex(double xp,double yp);
	char* HashListPtr(unsigned index);
	void SetHashListPtr(unsigned index,char *pPtr);
	char* HashListValuePtr(char* pList,unsigned value);
	void SetHashListValuePtr(char* pList,unsigned value,char *pPtr);

protected:
	// ----------
	// user input
	// ----------
	// parameters
	unsigned m_nulines, m_nvlines;	// number of u,v contour lines to draw
	double	 m_umin, m_umax;	   	// range of u parameter
	double	 m_vmin, m_vmax;	   	// range of v parameter

	// flags
	BOOL     m_draw_ulines;			// 0=no, 1=draw U lines
	BOOL     m_draw_vlines;			// 0=no, 1=draw V lines
	unsigned m_is_color_xyz;		// 0=u,v 1=color uses x,y,z

	// -------------
	// internal data
	// -------------
	double	m_maxdxp, m_maxdyp, m_maxdzp, m_zptol;	// max triangle deltas
	unsigned m_nPoints;		// number of points in m_pPoints
	unsigned m_nTriangles;	// number of triangles in m_pTriangles
	unsigned m_nutriangles, m_nvtriangles;
	Point3DT*	 m_pPoints;		// pointer to an array of Point3DT
	Triangle3DT* m_pTriangles;	// pointer to an array of Triiangle3DT

	// min/max projection values
	double m_xpmin,  m_xpmax;
	double m_ypmin,  m_ypmax;
	double m_zpmin,  m_zpmax;

	// hash look-up table for speed
	unsigned	m_nHashItems;	// number of items in m_pHashCounts and m_pHashPointers;
	unsigned*	m_pHashCounts;	// pointer to an array of counts
	char**		m_pHashPointers;// pointer to an array of pointers
	unsigned	m_nxbins;		// number of bins along x axis
	unsigned	m_nybins;		// number of bins along y axis
	double		m_dxbin;		// 1/(width of each bin)
	double		m_dybin;		// 1/(height of each bin)
};


#endif	/* if _PLOT3DT_H_ included */

/*	<><><><><><><><><><><><><>  Plot3DT.h  <><><><><><><><><><><><><><> */
